class DataSetConfigError(Exception):
    """``DataSetConfigError`` raised when dataset configuration is invalid or cannot be parsed. 
    """

    pass


class DataSetError(Exception):
    """``DataSetError`` raised when dataset operations fail.
    """

    pass
